/** X128: Spectrum 128 emulator ******************************/
/**                                                         **/
/**                       unix_all.c                        **/
/**                                                         **/
/** This file contains Unix/X-dependent subroutines and     **/
/** drivers.                                                **/
/**                                                         **/
/** Copyright (C) James McKay     1996                      **/
/**                                                         **/
/** Underlying X-Windows drivers from VGB & FMSX by:        **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995                 **/
/**               Elan Feingold   1995                      **/
/**               Arnold Metselaar1995                      **/
/**                                                         **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

/** MIT Shared Memory Extension for X ************************/
#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
XShmSegmentInfo SHMInfo;
int UseSHM=1;
#endif

#define ColTyp unsigned long
#define ColFac 9362 /* 2^16 / 7 */

char *Title="X128 Unix/X 0.5";
int ExitFlag=0;
int SaveCPU=0;
struct itimerval TimerValue;

/** Various X-related variables ******************************/
Screen *screen;
Display *disp;
Window root,win;
Colormap cmap;
XImage *ximage;
XEvent ev;
GC defaultGC,gc;
int screen_num,bitpix;
unsigned long white,black;
Bool lsbfirst;
UC Xtab[4];
UC *XBuf; /* XBuf! */

unsigned long *XBuf8;
UC XPal[16];
UC Scr_Msb_First=0;

#ifdef SCALE2
UC mono0[256][64];
UC mono1[256][64];
#else
UC mono[256][64];
#endif

/* size of window */
#ifdef SCALE2
	#define WIDTH 640
	#define WIDTHR2 160
	#define WIDTHR3 80
	#define HEIGHT 384
#else
	#define WIDTH 320
	#define WIDTHR2 80
	#define WIDTHR3 40
	#define HEIGHT 192
#endif

/* I assume that this is a 24-bit palette, and it is. */

char *ColorNames[16] = { "black","#0000A3","#DF0000","#E700B7", "#00D700","#00D7D7","#D3D300","#CBCBCB","black","#0000AF","#F30000","#FF00DF", "#00F300","#00FFFF","#FFFF00","#FFFFFF"};

/* Makes 16K monochrome data table so that mono screen is nice */
#define OTHER_MONO

word expand_char(UC scr_val)
{
	UC x=0;
	word scr_word=0;
	pair temp_word;

	temp_word.W=scr_val;
	for(x=0;x<8;x++)
	{
		temp_word.W<<=1;
		if(temp_word.B.h&1)
		{
			scr_word=(scr_word<<2)|3;
		}
		else
		{
			scr_word=(scr_word<<2);
		}			
	}
	return scr_word;
}

#ifdef OTHER_MONO
void construct_mono()
{
	UC scr_val, ink, paper;
	int attr, data, bit_p;
	pair temp;

	for(attr=0;attr<64;attr++)
	{
		temp.W=0;
		ink=attr&7;
		paper=(attr&56)>>3;
		for(data=0;data<256;data++)
		{
			scr_val=0;
			temp.B.h=data;
			for(bit_p=0;bit_p<8;bit_p++)
			{
				temp.W>>=1;
				if((temp.B.l&128)==128)
				{
					if(paper>3) /* White back */
	  				{
						if(ink!=paper)
						{
							if(Scr_Msb_First)
							{
								scr_val|=ormask[bit_p];
							}
							else
							{
								scr_val|=ormask[7-bit_p];
							}
						}
	  				}
					else
					{
						if(ink==paper)
						{
							if(Scr_Msb_First)
							{
								scr_val|=ormask[bit_p];
							}
							else
							{
								scr_val|=ormask[7-bit_p];
							}
						}
					}
				}
       				else
				{
					if(paper<4) /* Black back */
					{
						if(Scr_Msb_First)
						{
							scr_val|=ormask[bit_p];
						}
						else
						{
							scr_val|=ormask[7-bit_p];						}
					}
				}
			}
#ifdef SCALE2
			temp=expand_char(scr_val);
			mono0[data][attr]=temp.B.h;
			mono1[data][attr]=temp.B.l;
#else
			mono[data][attr]=scr_val;
#endif
		}
	}
}
#else
void construct_mono()
{
	UC scr_val;
	int attr, data, bit_p;
	pair temp;

	for(attr=0;attr<64;attr++)
	{
		temp.W=0;
		for(data=0;data<256;data++)
		{
			scr_val=0;
			temp.B.h=data;
			for(bit_p=0;bit_p<8;bit_p++)
			{
				temp.W>>=1;
				if((temp.B.l&128)==128)
				{
					if((attr&7)<4)
	  				{
						if(Scr_Msb_First)
						{
							scr_val|=ormask[bit_p];
						}
						else
						{
							scr_val|=ormask[7-bit_p];
						}
	  				}
				}
       				else
				{
					if((attr&56)<32)
					{
						if(Scr_Msb_First)
						{
							scr_val|=ormask[bit_p];
						}
						else
						{
							scr_val|=ormask[7-bit_p];
						}
					}
				}
			}
			mono[data][attr]=scr_val;
		}
	}
}
#endif

void OnBreak(int ARG)
{ 
	quit=QUIT_COMPLETELY;
	if (Verbose & 1) printf("Killed by signal %d.\n",ARG);
}

void OnTimer(int ARG)
{
	int_50=1;signal(SIGVTALRM,OnTimer);
}

/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  int J;

  if(Verbose) printf("Shutting down...\n");
  if(disp&&win)
  {
#ifdef MITSHM
    if(UseSHM)
    {
      XShmDetach(disp,&SHMInfo);
      if(SHMInfo.shmaddr) shmdt(SHMInfo.shmaddr);
      if(SHMInfo.shmid>=0) shmctl(SHMInfo.shmid,IPC_RMID,0);
    }
    else
#endif MITSHM
    if(ximage) XDestroyImage(ximage);
    XAutoRepeatOn(disp);
  }
  if(disp) XCloseDisplay(disp);
}

/* returns 1 if format X is 'better' than format Y, 0 otherwise */
int better(XPixmapFormatValues X, XPixmapFormatValues Y)
{ 
  if ((X.bits_per_pixel % 8)&&(8 % X.bits_per_pixel)) return 0; /*X legal ? */
  if ((X.depth > Y.depth) && (Y.depth < 8)) return 1;
  if ((X.bits_per_pixel < Y.bits_per_pixel) && (X.depth >=8)) return 1;
  return 0;
}

/** InitMachine **********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  XGCValues values;
  int J,K,L;
  Bool SharedPixmaps;
  int nform,i,depth; /* # pixmapformats, index, depth */
  XPixmapFormatValues *flist,choice; /* list of format describers and choice */
  ColTyp T;
  char *order;
  XColor Color,C;

  if(Verbose)
    printf("Initializing X11 drivers:\n  Opening display...");
  disp=XOpenDisplay(NULL);
  if(!disp) { if(Verbose) printf("FAILED\n");return(0); }

  screen=DefaultScreenOfDisplay(disp);screen_num=DefaultScreen(disp);
  white=XWhitePixel(disp,screen_num);black=XBlackPixel(disp,screen_num);
  root=DefaultRootWindow(disp);
  defaultGC=XDefaultGC(disp,screen_num);
  cmap=DefaultColormap(disp,screen_num);
#ifdef MITSHM
  if (UseSHM)
    if ( (UseSHM=XQueryExtension(disp,"MIT-SHM",&J,&K,&L)) )
    { 
      XShmQueryVersion(disp,&J,&K,&SharedPixmaps);
      UseSHM=SharedPixmaps;
    }
#endif

  if(Verbose) printf("OK\n  Choosing picture format...");
  flist=XListPixmapFormats(disp,&nform);
  choice.bits_per_pixel=choice.depth=0;
  for (i=0;i<nform;i++) /* search for 'best' format */
    if (better(flist[i],choice)) choice=flist[i];
  bitpix=choice.bits_per_pixel;
  depth =choice.depth; 
  XFree((void *)flist);
  if (!bitpix) { if(Verbose) printf("FAILED\n"); return(0); }
  order="";
  if (bitpix>8)
  {
    order=(ImageByteOrder(disp)==LSBFirst) ? "(LSB first)" : "(MSB first)";
    T = (ImageByteOrder(disp)==LSBFirst)? 0x03020100 : 0x00010203 ;

    for (i=0;i<4;i++) Xtab[*((UC *) &T+i)]=i;
    if ((ImageByteOrder(disp)==LSBFirst) && (bitpix<32))
      memcpy(Xtab+4-(bitpix/8), Xtab, bitpix/8);
  }
  if (bitpix<=8)
    order=(lsbfirst=(BitmapBitOrder(disp)==LSBFirst))
         ? "(lsb first)" : "(msb first)";
  if (BitmapBitOrder(disp)==LSBFirst)
  {
      Scr_Msb_First=0;
  }
  else
  {
      Scr_Msb_First=1;
  }
  if (Verbose)
    printf("depth=%d, %d bit%s per pixel %s\n",
           depth,bitpix,(bitpix==1)?"":"s",order);

  if(Verbose) printf("  Opening window...");
  win=XCreateWindow(disp,root,0,0,WIDTH,HEIGHT,0,depth,InputOutput,
                    DefaultVisual(disp,screen_num),0,NULL);
  if(!win) { if(Verbose) printf("FAILED\n");return(0); }

  {
    XSizeHints hints;
    XWMHints wmhints;

    hints.flags=USSize; hints.width=WIDTH; hints.height=HEIGHT;
    wmhints.input=True; wmhints.flags=InputHint;
    XSetWMHints(disp,win,&wmhints);
    XSetWMNormalHints(disp,win,&hints);
    XStoreName(disp,win,Title);
  }

  XSelectInput(disp,win,ExposureMask|KeyPressMask|KeyReleaseMask);
  XMapRaised(disp,win); XNextEvent(disp,&ev);

  gc=XCreateGC(disp,win,0,&values);
  XSetForeground(disp,gc,white);
  XSetBackground(disp,gc,black);
  XClearWindow(disp,win);

  XAutoRepeatOff(disp);

#ifdef MITSHM
  if(UseSHM)
  {
    if(Verbose)
      printf("OK\n  Using shared memory:\n    Creating image...");
    ximage=XShmCreateImage
           (
             disp,DefaultVisual(disp,screen_num),depth,
             ZPixmap,NULL,&SHMInfo,WIDTH,HEIGHT
           );
    if(!ximage) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Getting SHM info...");
    SHMInfo.shmid=shmget
                  (
                    IPC_PRIVATE,ximage->bytes_per_line*ximage->height,
                    IPC_CREAT|0777
                  );
    if(SHMInfo.shmid<0) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Allocating SHM...");
if(bitpix==1)
{
    XBuf=(UC *)(ximage->data=SHMInfo.shmaddr=shmat(SHMInfo.shmid,0,0));
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }
}
else
{
    XBuf8=(unsigned long *)(ximage->data=SHMInfo.shmaddr=shmat(SHMInfo.shmid,0,0));
    if(!XBuf8) { if(Verbose) printf("FAILED\n");return(0); }
}

    SHMInfo.readOnly=False;
    if(Verbose) printf("OK\n    Attaching SHM...");
    if(!XShmAttach(disp,&SHMInfo))
      { if(Verbose) printf("FAILED\n");return(0); }
  }
  else
#endif
  {
    if(Verbose) printf("OK\n  Allocating RAM...");
if(bitpix==1)
{
    XBuf=(UC *)malloc(sizeof(UC)*HEIGHT*WIDTH*bitpix/8);
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }
}
else
{
    XBuf8=(unsigned long *)malloc(sizeof(UC)*HEIGHT*WIDTH*bitpix/8);
    if(!XBuf8) { if(Verbose) printf("FAILED\n");return(0); }
}
if(bitpix==1)
{
    if(Verbose) printf("OK\n  Creating image...");
    ximage=XCreateImage
           (
             disp,DefaultVisual(disp,screen_num),depth,
             ZPixmap,0,XBuf,WIDTH,HEIGHT,8,0
           );
    if(!ximage) { if(Verbose) printf("FAILED\n");return(0); }
}
else
{
    if(Verbose) printf("OK\n  Creating image...");
    ximage=XCreateImage
           (
             disp,DefaultVisual(disp,screen_num),depth,
             ZPixmap,0,(char *)XBuf8,WIDTH,HEIGHT,8,0
           );
    if(!ximage) { if(Verbose) printf("FAILED\n");return(0); }
}
  }
  if(Verbose) printf("OK\n");

  XPal[0]=
    XAllocNamedColor(disp,cmap,ColorNames[0],&Color,&C)? Color.pixel:black;
  XPal[1]=
    XAllocNamedColor(disp,cmap,ColorNames[1],&Color,&C)? Color.pixel:black;
  XPal[2]=
    XAllocNamedColor(disp,cmap,ColorNames[2],&Color,&C)? Color.pixel:white;
  XPal[3]=
    XAllocNamedColor(disp,cmap,ColorNames[3],&Color,&C)? Color.pixel:white;
  XPal[4]=
    XAllocNamedColor(disp,cmap,ColorNames[4],&Color,&C)? Color.pixel:black;
  XPal[5]=
    XAllocNamedColor(disp,cmap,ColorNames[5],&Color,&C)? Color.pixel:black;
  XPal[6]=
    XAllocNamedColor(disp,cmap,ColorNames[6],&Color,&C)? Color.pixel:white;
  XPal[7]=
    XAllocNamedColor(disp,cmap,ColorNames[7],&Color,&C)? Color.pixel:white;
  XPal[8]=
    XAllocNamedColor(disp,cmap,ColorNames[8],&Color,&C)? Color.pixel:black;
  XPal[9]=
    XAllocNamedColor(disp,cmap,ColorNames[9],&Color,&C)? Color.pixel:black;
  XPal[10]=
    XAllocNamedColor(disp,cmap,ColorNames[10],&Color,&C)? Color.pixel:white;
  XPal[11]=
    XAllocNamedColor(disp,cmap,ColorNames[11],&Color,&C)? Color.pixel:white;
  XPal[12]=
    XAllocNamedColor(disp,cmap,ColorNames[12],&Color,&C)? Color.pixel:black;
  XPal[13]=
    XAllocNamedColor(disp,cmap,ColorNames[13],&Color,&C)? Color.pixel:black;
  XPal[14]=
    XAllocNamedColor(disp,cmap,ColorNames[14],&Color,&C)? Color.pixel:white;
  XPal[15]=
    XAllocNamedColor(disp,cmap,ColorNames[15],&Color,&C)? Color.pixel:white;

  /*if(Verbose&1) printf("Initializing timer for 50Hz...");
  signal(SIGVTALRM,OnTimer);
  TimerValue.it_interval.tv_sec=
  TimerValue.it_value.tv_sec=0L;
  TimerValue.it_interval.tv_usec=
  TimerValue.it_value.tv_usec=1000000L/50;
  J=setitimer(ITIMER_VIRTUAL,&TimerValue,NULL);
  if(Verbose&1) printf(J ? "FAILED\n":"OK\n");
  if(J) return (0);*/

  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  if(bitpix==1)
  {
	construct_mono();
  }
  return(1);
}

/** PutImage *************************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void PutImage()
{
#ifdef MITSHM
  if(UseSHM)
    XShmPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT,False);
  else
#endif
  XPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT);
  XFlush(disp);
}

/** Keyboard *************************************************/
/** Return the current keyboard state.                      **/
/*************************************************************/
void Keyboard(void)
{
  static UC c_shift = 0;
  word J;

  if(XCheckWindowEvent(disp,win,KeyPressMask|KeyReleaseMask,&ev))
  {
    J=XLookupKeysym((XKeyEvent *)&ev,0);
    if(ev.type==KeyPress)
      switch(J)
      {
	case XK_F1: case XK_Help: quit=QUIT_HELP;break;/* F1 */
	case XK_F2: quit=QUIT_NMI;break; /* F2 */
	case XK_F3: quit=QUIT_RESET;break;  /* F3 */
	case XK_F4: quit=QUIT_JOY;break; /* F4 */
	case XK_F5: quit=QUIT_LOAD;break;/* F5 */
	case XK_F6: quit=QUIT_SAVE;break;/* F6 */
	case XK_F7: quit=QUIT_TAP;break; /* F7 */
	case XK_F8:     if(c_shift)
			{
				if(ULA_delay<255) ULA_delay++;
			}
			else
			{
				quit=QUIT_VOC;
			}
			break;
	case XK_F9:     if(c_shift)
			{
				if(ULA_delay) ULA_delay--;
			}
			break;
	case XK_F10: quit=QUIT_COMPLETELY;break;
	case XK_F11: quit=QUIT_SOUND;break;

	case XK_Escape: SKey[CtoV]&=on0;SKey[N1toN5]&=on0;break; /* Esc,Edit */
        case XK_Tab:    switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]&=on0;
					SKey[N0toN6]&=on0;break; /* Tab,Fire */
				case 1 : In31|=off4;break; /* Kempston Fire */
				case 2 : SKey[N1toN5]&=on4;break; /* 5 */
				case 3 : SKey[N0toN6]&=on0;break; /* 0 */
			 }
			 break;
        case XK_Down:   switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]&=on0;
					SKey[N0toN6]&=on4;break; /* Down Arrow */
				case 1 : In31|=off2;break; /* Kempston Down */
				case 2 : SKey[N1toN5]&=on2;break; /* 3 */
				case 3 : SKey[N0toN6]&=on2;break; /* 8 */
			 }
			 break;
        case XK_Up: 	switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]&=on0;
					SKey[N0toN6]&=on3;break; /* Up Arrow */
				case 1 : In31|=off3;break; /* Kempston Up */
				case 2 : SKey[N1toN5]&=on3;break; /* 4 */
				case 3 : SKey[N0toN6]&=on1;break; /* 9 */
			 }
			 break;
        case XK_Left:   switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]&=on0;
					SKey[N1toN5]&=on4;break; /* Left Arrow */
				case 1 : In31|=off1;break; /* Kempston Left */
				case 2 : SKey[N1toN5]&=on0;break; /* 1 */
				case 3 : SKey[N0toN6]&=on4;break; /* 6 */
			 }
			 break;
        case XK_Right:  switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]&=on0;
					SKey[N0toN6]&=on2;break; /* Right Arrow */
				case 1 : In31|=off0;break; /* Kempston Right */
				case 2 : SKey[N1toN5]&=on1;break; /* 2 */
				case 3 : SKey[N0toN6]&=on3;break; /* 7 */
			 }
			 break;
	case XK_BackSpace:SKey[CtoV]&=on0;SKey[N0toN6]&=on0;break;
	case XK_Delete: SKey[CtoV]&=on0;SKey[N0toN6]&=on0;break;

        case XK_p:	SKey[PtoY]&=on0;break;
        case XK_o:	SKey[PtoY]&=on1;break;
        case XK_i:	SKey[PtoY]&=on2;break;
        case XK_u:	SKey[PtoY]&=on3;break;
        case XK_y:	SKey[PtoY]&=on4;break;

        case XK_q:	SKey[QtoT]&=on0;break;
        case XK_w:	SKey[QtoT]&=on1;break;
        case XK_e:	SKey[QtoT]&=on2;break;
        case XK_r:	SKey[QtoT]&=on3;break;
        case XK_t:	SKey[QtoT]&=on4;break;

        case XK_0:	SKey[N0toN6]&=on0;break;
        case XK_9:	SKey[N0toN6]&=on1;break;
        case XK_8:	SKey[N0toN6]&=on2;break;
        case XK_7:	SKey[N0toN6]&=on3;break;
        case XK_6:	SKey[N0toN6]&=on4;break;

        case XK_1:	SKey[N1toN5]&=on0;break;
        case XK_2:	SKey[N1toN5]&=on1;break;
        case XK_3:	SKey[N1toN5]&=on2;break;
        case XK_4:	SKey[N1toN5]&=on3;break;
        case XK_5:	SKey[N1toN5]&=on4;break;

        case XK_a:	SKey[AtoG]&=on0;break;
        case XK_s:	SKey[AtoG]&=on1;break;
        case XK_d:	SKey[AtoG]&=on2;break;
        case XK_f:	SKey[AtoG]&=on3;break;
        case XK_g:	SKey[AtoG]&=on4;break;

        case XK_Shift_L:SKey[CtoV]&=on0;c_shift=1;break;
        case XK_Shift_R:SKey[CtoV]&=on0;c_shift=1;break;

        case XK_z:	SKey[CtoV]&=on1;break;
        case XK_x:	SKey[CtoV]&=on2;break;
        case XK_c:	SKey[CtoV]&=on3;break;
        case XK_v:	SKey[CtoV]&=on4;break;

        case XK_Return:SKey[EtoH]&=on0;break;
        case XK_l:	SKey[EtoH]&=on1;break;
        case XK_k:	SKey[EtoH]&=on2;break;
        case XK_j:	SKey[EtoH]&=on3;break;
        case XK_h:	SKey[EtoH]&=on4;break;

        case XK_space:SKey[StoB]&=on0;break;
        case XK_Alt_L:SKey[StoB]&=on1;break;
        case XK_Alt_R:SKey[StoB]&=on1;break;
        case XK_Meta_L:SKey[StoB]&=on1;break;
        case XK_Meta_R:SKey[StoB]&=on1;break;

        case XK_m:	SKey[StoB]&=on2;break;
        case XK_n:	SKey[StoB]&=on3;break;
        case XK_b:	SKey[StoB]&=on4;break;

	case XK_minus:  if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]&=on1;SKey[N0toN6]&=on0;/* Underline*/
			 }
			 else
			 {
				SKey[StoB]&=on1;SKey[EtoH]&=on3;  /* Minus */
			 }
			 break;
	case XK_equal: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]&=on1;SKey[EtoH]&=on2;  /* Plus */
			 }
			 else
			 {
				SKey[StoB]&=on1;SKey[EtoH]&=on1;  /* Equals */
			 }
			 break;
	case XK_semicolon: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]&=on1;SKey[CtoV]&=on1;/* Colon */
			 }
			 else
			 {
				SKey[StoB]&=on1;SKey[PtoY]&=on1;  /* Semi Colon */
			 }
			 break;
	case XK_apostrophe: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]&=on1;SKey[PtoY]&=on0;  /* Inverted Comma */
			 }
			 else
			 {
				SKey[StoB]&=on1;SKey[N0toN6]&=on3;  /* Apostrophe */
			 }
			 break;
	case XK_comma: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]&=on1;SKey[QtoT]&=on3;/* Lesser Than */
			 }
			 else
			 {
				SKey[StoB]&=on1;SKey[StoB]&=on3;  /* Comma */
			 }
			 break;
	case XK_period: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]&=on1;SKey[QtoT]&=on4;  /* Greater Than */
			 }
			 else
			 {
				SKey[StoB]&=on1;SKey[StoB]&=on2;  /* Full Stop */
			 }
			 break;
	case XK_slash: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]&=on1;SKey[CtoV]&=on3;  /* Query */
			 }
			 else
			 {
				SKey[StoB]&=on1;SKey[CtoV]&=on4;  /* Slash */
			 }
			 break;

	case XK_Control_L: SKey[CtoV]&=on0;SKey[StoB]&=on1;break;/*Extend*/
	case XK_Control_R: SKey[CtoV]&=on0;SKey[StoB]&=on1;break;/*Extend*/
      }
    else
     switch(J)
      {
	case XK_Escape:SKey[CtoV]|=off0;SKey[N1toN5]|=off0;break; /* Esc,Edit */
        case XK_Tab:    switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]|=off0;
					SKey[N0toN6]|=off0;break; /* Tab,Fire */
				case 1 : In31&=on4;break; /* Kempston Fire */
				case 2 : SKey[N1toN5]|=off4;break; /* 5 */
				case 3 : SKey[N0toN6]|=off0;break; /* 0 */
			 }
			 break;
        case XK_Down:   switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]|=off0;
					SKey[N0toN6]|=off4;break; /* Down Arrow */
				case 1 : In31&=on2;break; /* Kempston Down */
				case 2 : SKey[N1toN5]|=off2;break; /* 3 */
				case 3 : SKey[N0toN6]|=off2;break; /* 8 */
			 }
			 break;
        case XK_Up:     switch(joy_type)
			 {
				case 0 : if(curs_caps) SKey[CtoV]|=off0;
					SKey[N0toN6]|=off3;break; /* Up Arrow */
				case 1 : In31&=on3;break; /* Kempston Up */
				case 2 : SKey[N1toN5]|=off3;break; /* 4 */
				case 3 : SKey[N0toN6]|=off1;break; /* 9 */
			 }
			 break;
        case XK_Left:   switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]|=off0;
					SKey[N1toN5]|=off4;break; /* Left Arrow */
				case 1 : In31&=on1;break; /* Kempston Left */
				case 2 : SKey[N1toN5]|=off0;break; /* 1 */
				case 3 : SKey[N0toN6]|=off4;break; /* 6 */
			 }
			 break;
        case XK_Right:  switch(joy_type)
			 {
				case 0 : 
					if(curs_caps) SKey[CtoV]|=off0;
					SKey[N0toN6]|=off2;break; /* Right Arrow */
				case 1 : In31&=on0;break; /* Kempston Right */
				case 2 : SKey[N1toN5]|=off1;break; /* 2 */
				case 3 : SKey[N0toN6]|=off3;break; /* 7 */
			 }
			 break;
	case XK_BackSpace:SKey[CtoV]|=off0;SKey[N0toN6]|=off0;break;
	case XK_Delete: SKey[CtoV]|=off0;SKey[N0toN6]|=off0;break;

        case XK_p:	SKey[PtoY]|=off0;break;
        case XK_o:	SKey[PtoY]|=off1;break;
        case XK_i:	SKey[PtoY]|=off2;break;
        case XK_u:	SKey[PtoY]|=off3;break;
        case XK_y:	SKey[PtoY]|=off4;break;

        case XK_q:	SKey[QtoT]|=off0;break;
        case XK_w:	SKey[QtoT]|=off1;break;
        case XK_e:	SKey[QtoT]|=off2;break;
        case XK_r:	SKey[QtoT]|=off3;break;
        case XK_t:	SKey[QtoT]|=off4;break;

        case XK_0:	SKey[N0toN6]|=off0;break;
        case XK_9:	SKey[N0toN6]|=off1;break;
        case XK_8:	SKey[N0toN6]|=off2;break;
        case XK_7:	SKey[N0toN6]|=off3;break;
        case XK_6:	SKey[N0toN6]|=off4;break;

        case XK_1:	SKey[N1toN5]|=off0;break;
        case XK_2:	SKey[N1toN5]|=off1;break;
        case XK_3:	SKey[N1toN5]|=off2;break;
        case XK_4:	SKey[N1toN5]|=off3;break;
        case XK_5:	SKey[N1toN5]|=off4;break;

        case XK_a:	SKey[AtoG]|=off0;break;
        case XK_s:	SKey[AtoG]|=off1;break;
        case XK_d:	SKey[AtoG]|=off2;break;
        case XK_f:	SKey[AtoG]|=off3;break;
        case XK_g:	SKey[AtoG]|=off4;break;

        case XK_Shift_L:SKey[CtoV]|=off0;c_shift=0;break;
        case XK_Shift_R:SKey[CtoV]|=off0;c_shift=0;break;

        case XK_z:	SKey[CtoV]|=off1;break;
        case XK_x:	SKey[CtoV]|=off2;break;
        case XK_c:	SKey[CtoV]|=off3;break;
        case XK_v:	SKey[CtoV]|=off4;break;

        case XK_Return:SKey[EtoH]|=off0;break;
        case XK_l:	SKey[EtoH]|=off1;break;
        case XK_k:	SKey[EtoH]|=off2;break;
        case XK_j:	SKey[EtoH]|=off3;break;
        case XK_h:	SKey[EtoH]|=off4;break;

        case XK_space:SKey[StoB]|=off0;break;
        case XK_Alt_L:SKey[StoB]|=off1;break;
        case XK_Alt_R:SKey[StoB]|=off1;break;
        case XK_Meta_L:SKey[StoB]|=off1;break;
        case XK_Meta_R:SKey[StoB]|=off1;break;

        case XK_m:	SKey[StoB]|=off2;break;
        case XK_n:	SKey[StoB]|=off3;break;
        case XK_b:	SKey[StoB]|=off4;break;

	case XK_minus:  if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]|=off1;SKey[N0toN6]|=off0;SKey[CtoV]&=on0;/* Underline*/
			 }
			 else
			 {
				SKey[StoB]|=off1;SKey[EtoH]|=off3;  /* Minus */
			 }
			 break;
	case XK_equal: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]|=off1;SKey[EtoH]|=off2;SKey[CtoV]&=on0;  /* Plus */
			 }
			 else
			 {
				SKey[StoB]|=off1;SKey[EtoH]|=off1;  /* Equals */
			 }
			 break;
	case XK_semicolon: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]|=off1;SKey[CtoV]|=off1;SKey[CtoV]&=on0; /* Colon */
			 }
			 else
			 {
				SKey[StoB]|=off1;SKey[PtoY]|=off1;  /* Semi Colon */
			 }
			 break;
	case XK_apostrophe: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]|=off1;SKey[PtoY]|=off0;SKey[CtoV]&=on0;  /* Inverted Comma */
			 }
			 else
			 {
				SKey[StoB]|=off1;SKey[N0toN6]|=off3;  /* Apostrophe */
			 }
			 break;
	case XK_comma: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]|=off1;SKey[QtoT]|=off3;SKey[CtoV]&=on0; /* Lesser Than */
			 }
			 else
			 {
				SKey[StoB]|=off1;SKey[StoB]|=off3;  /* Comma */
			 }
			 break;
	case XK_period: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]|=off1;SKey[QtoT]|=off4;SKey[CtoV]&=on0; /* Greater Than */
			 }
			 else
			 {
				SKey[StoB]|=off1;SKey[StoB]|=off2;  /* Full Stop */
			 }
			 break;
	case XK_slash: if(c_shift)
			 {
				SKey[CtoV]|=off0;SKey[StoB]|=off1;SKey[CtoV]|=off3;SKey[CtoV]&=on0; /* Query */
			 }
			 else
			 {
				SKey[StoB]|=off1;SKey[CtoV]|=off4;  /* Slash */
			 }
			 break;
	case XK_Control_L: SKey[CtoV]|=off0;SKey[StoB]|=off1;break;/*Extend*/
	case XK_Control_R: SKey[CtoV]|=off0;SKey[StoB]|=off1;break;/*Extend*/
      }
  }

  /*for(J=0;XCheckWindowEvent(disp,win,FocusChangeMask,&ev);)
    J=(ev.type==FocusOut);
  if(SaveCPU&&J)
  {
    XAutoRepeatOn(disp);	
    while(!XCheckWindowEvent(disp,win,FocusChangeMask,&ev)&&!ExitFlag)
    {
      if(XCheckWindowEvent(disp,win,ExposureMask,&ev)) PutImage();
      XPeekEvent(disp,&ev);
    }
    XAutoRepeatOff(disp);
  }*/
}

void init_keyboard(void)
{
	XAutoRepeatOff(disp);
}

void release_keyboard(void)
{
	XAutoRepeatOn(disp);
}

#include "common.h"
