/******************************************/
/**                                      **/
/**        X128 V0.4 Unix/X Version      **/
/**                                      **/
/**          (C) James McKay 1996        **/
/**                                      **/
/**    This software may not be used     **/
/**    for commercial reasons, the code  **/
/**    may not be modified or reused     **/
/**    without permission.               **/
/**                                      **/
/******************************************/
/**                                      **/
/** Special Guest Stars :                **/
/**                                      **/
/**  Marat Fayzullin 1995 (Unix & Ideas) **/
/** Arnold Metselaar 1995 (More Unix)    **/
/******************************************/

#define MAIN_OBJ

/* #define SCALE2 */
#define MITSHM
#define Verbose 1

/* Crucial!!!! */
#include "x128_end.c" /* Has LMB/MSB info. */

#include "x128_var.c"

#include "ay8910.h"

/* Prototypes */

void update_8bitline(UC linenum);
void update_monoline(UC linenum);
void clean_regs(void);
UC return_next_VOC_bit(void); /* In x128_voc.c */
void RefreshScreen(void);
void save_PSG(void);

#include "unix.c"

/* Code */

void Sound_on(void)
{
	if(all_sound_on)
	{
		if(!model_48)
		{
			if(adlib_present)
			{
				ResumeSound();
			}
		}
		/* No need to switch on internal */
	}
}

void Sound_off(void)
{
	if(adlib_present)
	{
		MuteSound();
	}
#ifdef LINUX
	outp(0x61,inp(0x61)&0xFC);
#endif
}

void RePaintMono(void)
{
	UC y;

	for(y=0;y<192;y++)
	{
		update_monoline(y);
	}
	RefreshScreen();
}

void RePaint8Bit(void)
{
	UC y;

	for(y=0;y<192;y++)
	{
		update_8bitline(y);
	}
	RefreshScreen();
}

void RePaintScreen(void)
{
	if(bitpix==8)
	{
		RePaint8Bit();
	}
	else
	{
		RePaintMono();
	}
}

#include "x128_def.c"

/* Very Non Portable FSL Stuff */

#define FNAME_SIZE 20 /* Only how much is shown */
#define UNIT_SIZE 260
#define FSL_WIN_WIDTH 24
#define FSL_WIN_START 4
#define FSL_BUFF UNIT_SIZE*2048

/* Port Read Keycodes For Input String */

#define X128_ESCAPE XK_Escape
#define X128_UP XK_Up
#define X128_DOWN XK_Down
#define X128_LEFT XK_Left
#define X128_RIGHT XK_Right
#define X128_RETURN XK_Return
#define X128_PAGE_UP XK_Prior
#define X128_PAGE_DOWN XK_Next
#define X128_KP_MINUS XK_KP_Subtract
#define X128_KP_PLUS XK_KP_Add
#define X128_HOME XK_Home
#define X128_END XK_End
#define X128_INSERT XK_Insert
#define X128_DELETE XK_Delete
#define X128_BACKSPACE XK_BackSpace

word fsl_asciiext_key(void)
{
  word J;
  static word loopy=0;

  while(1)
  {
    if(XCheckWindowEvent(disp,win,KeyPressMask|KeyReleaseMask,&ev))
    {
      J=XLookupKeysym((XKeyEvent *)&ev,0);
      if(ev.type==KeyPress)
	return J;
    }
    if(!++loopy)
    {
	RePaintScreen();
    }
  }
}

word fsl_extend_key(void)
{
  word J;
  static word loopy=0;

  while(1)
  {
    if(XCheckWindowEvent(disp,win,KeyPressMask|KeyReleaseMask,&ev))
    {
      J=XLookupKeysym((XKeyEvent *)&ev,0);
      if(ev.type==KeyPress)
	return J;
    }
    if(!++loopy)
    {
	RePaintScreen();
    }
  }
}

#include "x128_fsl.c"
#include "x128_tap.c"
#include "x128_voc.c"
#include "x128_cb.h"
#include "x128_ed.h"
#include "x128_z80.h"

void make_table(void)
{
	int v,x,y,z;
	pair temp;
	quadruple qtemp;

	for (y=0;y<192;y++)
	{
		vga_x_y[y]=(y+4)*80+8;
	}
	for (x=0;x<8;x++)
	{
		qtemp.B.b1=XPal[x];
		qtemp.B.b2=XPal[x];
		qtemp.B.b3=XPal[x];
		qtemp.B.b4=XPal[x];
		ul_border4[x]=qtemp.Q;
	}
	v=0;
	for (x=0;x<6144;x+=2048)
	{
		for (y=0;y<256;y+=32)
		{
			for (z=0;z<2048;z+=256)
			{
				spline[v]=x+y+z;
				v++;
			}
		}
	}
	for (x=0;x<256;x++)
	{
		y=x&7; /* fore */
		z=(x&56)>>3; /* back */
		if (!y) y|=8;
		if (!z) z|=8;
		if (x&64)
		{
			y=y|8;
			z=z|8;
		}
		fore_attrib[x]=y;
		back_attrib[x]=z;
	}

	for (x=0;x<16;x++)
	{
		for (y=0;y<128;y++)
		{
#ifndef LSB_FIRST
			if(x&8)
			{
				qtemp.B.b4=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b4=XPal[back_attrib[y]];
			}
			if(x&4)
			{
				qtemp.B.b3=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b3=XPal[back_attrib[y]];
			}
			if(x&2)
			{
				qtemp.B.b2=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b2=XPal[back_attrib[y]];
			}
			if(x&1)
			{
				qtemp.B.b1=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b1=XPal[back_attrib[y]];
			}
#else
			if(x&1)
			{
				qtemp.B.b4=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b4=XPal[back_attrib[y]];
			}
			if(x&2)
			{
				qtemp.B.b3=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b3=XPal[back_attrib[y]];
			}
			if(x&4)
			{
				qtemp.B.b2=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b2=XPal[back_attrib[y]];
			}
			if(x&8)
			{
				qtemp.B.b1=XPal[fore_attrib[y]];
			}
			else
			{
				qtemp.B.b1=XPal[back_attrib[y]];
			}
#endif
#ifdef SCALE2
			expand_long(qtemp.Q,&vga_h_nib0[x][y],&vga_h_nib1[x][y]);
#else
			vga_h_nib[x][y]=qtemp.Q;
#endif
		}
	}
	for (x=0;x<128;x++)
	{
		y=(x&7)<<3;
		z=(x&56)>>3;
		reverse_attr[x]=(x&64)|y|z;
	}
	for (x=0;x<256;x++)
	{
		calc_parity(x);
		parity[x]=(af.B.l&4);
	}
	for (x=0;x<8;x++)
	{
		SKey[x]=255;
	}
}

/* vga_line_0 and vga_line_1 have moved to common.h */

void parse_params(int argc, char *param[])
{
	char temp_str[MAXPATH];
	int arg_num;
	UC file_valid;

	if(quit==QUIT_COMPLETELY) return; /* Rom missing */
	arg_num=1;
	file_valid=0;
	while(arg_num<argc)
	{
		if((param[arg_num][0]=='-')||(param[arg_num][0]=='/'))
		{
			strcpy(temp_str,(char *)param[arg_num]+1);
			if(!(strcmp(temp_str,"?")))
			{
				printf("x128 V0.4 By James McKay\n");
				printf("\nOptions:\n");
				printf("/?     - This help.\n");
				printf("/128   - 128K mode.\n");
				printf("/48    - 48K mode.\n");
				printf("/quiet - Sound is off.\n");
				printf("/delay <number> - Set delay level");
				printf(" (0-65535)\n");
				printf("/frame <number> - Set frame skip");
				printf(" (0-255)\n");
				printf("/ula <number>   - Set ULA delay");
				printf(" (0-255)\n");
				printf("<filename> If the file has a .z80");
				printf(" or a .sna extension it");
				printf(" will be loaded.\n");
				printf("If it is a .tap or .voc it is");
				printf(" selected.\n");
				printf("\nPress F1 when in the emulator");
				printf(" for the help screen.\n");
				release_mem();
				exit(1);
			}
			if(!(strcmp(temp_str,"128")))
			{
				if(file_valid!=1)
				{
					quit=QUIT_RESET;
					skip_reset=0;
				}
			}
			if(!(strcmp(temp_str,"48")))
			{
				if(file_valid!=1)
				{
					quit=QUIT_RESET;
					skip_reset=1;
				}
			}
			if(!(strcmp(temp_str,"delay")))
			{
				arg_num++;
				slow_down=atoi(param[arg_num]);
			}
			if(!(strcmp(temp_str,"DELAY")))
			{
				arg_num++;
				slow_down=atoi(param[arg_num]);
			}
			if(!(strcmp(temp_str,"frame")))
			{
				arg_num++;
				frame_skip=atoi(param[arg_num]);
			}
			if(!(strcmp(temp_str,"FRAME")))
			{
				arg_num++;
				frame_skip=atoi(param[arg_num]);
			}
			if(!(strcmp(temp_str,"ula")))
			{
				arg_num++;
				ULA_delay=atoi(param[arg_num]);
			}
			if(!(strcmp(temp_str,"ULA")))
			{
				arg_num++;
				ULA_delay=atoi(param[arg_num]);
			}
			if(!(strcmp(temp_str,"quiet")))
			{
				all_sound_on=UseAdlib=0;
			}
			if(!(strcmp(temp_str,"QUIET")))
			{
				all_sound_on=UseAdlib=0;
			}
		}
		else
		{
			strcpy(temp_str,param[arg_num]);
			if(!different_extension(".z80",temp_str))
			{
				strcpy(the_path,temp_str);
				if(!(open_z80()))
				{
					halt=0; /* Re-enable processor for new snap */
				}
				file_valid=1;
				quit=DONT_QUIT;
			}
			if(!different_extension(".slt",temp_str))
			{
				strcpy(the_path,temp_str);
				if(!(open_z80()))
				{
					halt=0; /* Re-enable processor for new snap */
				}
				file_valid=1;
				quit=DONT_QUIT;
			}
			if(!different_extension(".sna",temp_str))
			{
				strcpy(the_path,temp_str);
				if(!(open_sna()))
				{
					halt=0;
					ED[77].opcode(); /* reti */
				}
				file_valid=1;
				quit=DONT_QUIT;
			}
			if(!different_extension(".tap",temp_str))
			{
				printf("TAP files are only selected,");
				printf(" not automatically loaded.\n");
				strcpy(the_path,temp_str);
				TAP_select();
				TAP_rompokes();
				file_valid=2;
			}
			if(!different_extension(".voc",temp_str))
			{
				printf("VOC files are only selected,");
				printf(" not automatically loaded.\n");
				strcpy(the_path,temp_str);
				open_VOC_input_file();
				undo_TAP_rompokes();
				file_valid=2;
			}
			if(!file_valid)
			{
				printf("File did not have z80, sna,");
				printf(" tap, voc or slt");
				printf(" extension, so I did not know\n");
				printf("how to deal with it.\n");
			}
		}
		arg_num++;
	}
	if(file_valid==1) skip_reset=255;
}

#ifdef OVERSCAN

#define main_bit(scr_drive)   \
  register UC byte;             \
  register UC screen_up=0;        \
  UC ff=0xFF;                      \
				    \
  do{                                 \
	if(t_state>=t_states_per_line)  \
	{                                 \
		t_state-=t_states_per_line; \
		line_counter++;              \
		if(++vline==s_lines)          \
		{                               \
			vline=0;                  \
			if(++screen_up==frame_skip) \
			{                             \
				RefreshScreen();       \
				screen_up=0;            \
			}                                 \
			Keyboard();                        \
			if(!--flash_count)                  \
			{                                     \
				flash_count=16;                 \
				flash_status^=1;                  \
			}                                           \
			if(iff1)                                    \
			{                                           \
				iff2=iff1;                          \
				iff1=halt=0;                        \
				if(im==2)                           \
				{                                   \
					push(pc.W);                 \
					ptemp.W=(ir.B.h<<8)|255;    \
					pc.W=wordpeek(ptemp.W);     \
					INC_R_REG;                  \
					t_state+=19;                \
				}                                   \
				else                                \
				{                                   \
					INC_R_REG;                  \
					rst_56();                   \
					t_state+=2;                 \
				}                                   \
			}                                           \
			if(PSG_open)                                \
			{                                           \
				fwrite(&ff,1,1,p_handle);           \
			}                                           \
		}                                                   \
		if((!screen_up)&&(vline&192))                       \
		{                                                   \
			scr_drive(vline-64);                        \
			t_state+=ULA_delay;                         \
		}                                                   \
	}                                                           \
	if(!halt)                                                   \
	{                                                           \
		INC_R_REG;                                          \
		byte=speekb(pc.W);                                  \
		pc.W++;                                             \
		Z80[byte].opcode();                                 \
	}                                                           \
	else                                                        \
	{                                                           \
		Z80[0].opcode();                                    \
	}

#else

#define main_bit(scr_drive)  \
  register UC byte;           \
  register UC screen_up=0;     \
  UC ff=0xFF;                   \
                                 \
  do{                             \
	if(int_50)                 \
	{                           \
		int_50=0;            \
		if(++screen_up==frame_skip) \
		{                          \
			screen_up=0;        \
			scr_drive();         \
		}                             \
		if(!--flash_count)             \
		{                              \
			flash_count=16;        \
			flash_status^=1;       \
		}                              \
		if(iff1)                       \
		{                              \
			iff2=iff1;             \
			iff1=halt=0;           \
			if(im==2)              \
			{                      \
				push(pc.W);    \
				ptemp.W=(ir.B.h<<8)|255; \
				pc.W=wordpeek(ptemp.W);  \
				INC_R_REG;               \
			}                                \
			else                             \
			{                                \
				INC_R_REG;               \
				rst_56;                  \
			}                                \
		}                                        \
		if(PSG_open)                             \
		{                                        \
			fwrite(&ff,1,1,p_handle);        \
		}                                        \
	}                                                \
	if(!halt)                                        \
	{                                                \
		INC_R_REG;                               \
		byte=speekb(pc.W);                       \
		pc.W++;                                  \
		Z80[byte].opcode();                      \
	}                                                \
	else                                             \
	{                                                \
		Z80[0].opcode();                         \
	}

#endif

#define main_bit_2 \
  }while (!quit);

void main_8bit_loop_delay(void)
{
#ifdef OVERSCAN
	main_bit(update_8bitline);
#else
	main_bit(RePaint8Bit);
#endif
	for(ptemp.W=0;ptemp.W<slow_down;ptemp.W++)
	{
		qtemp.Q+=2;
		qtemp.Q-=2;
	};
	main_bit_2;
}
void main_8bit_loop(void)
{
#ifdef OVERSCAN
	main_bit(update_8bitline);
#else
	main_bit(RePaint8Bit);
#endif
	main_bit_2;
}

void main_mono_loop_delay(void)
{
#ifdef OVERSCAN
	main_bit(update_monoline);
#else
	main_bit(RePaintMono);
#endif
	for(ptemp.W=0;ptemp.W<slow_down;ptemp.W++)
	{
		qtemp.Q+=2;
		qtemp.Q-=2;
	}
	main_bit_2;
}

void main_mono_loop(void)
{
#ifdef OVERSCAN
	main_bit(update_monoline);
#else
	main_bit(RePaintMono);
#endif
	main_bit_2;
}

void main(int argc, char *argv[])
{
  char *temp_str, *getcwd();
  UC ULA_param=0;
  UC discard;

  error_str="No error.";

  temp_str=getcwd((char *)NULL,MAXPATH);
  strcat(temp_str,"/");
  strcat(temp_str,"x128");
  printf("temp_str=%s\n",temp_str);
  strcpy(the_sna_path,temp_str);
  strcpy(the_tap_path,temp_str);
  strcpy(the_dat_path,temp_str);
  strcpy(the_psg_path,temp_str);
  strcpy(the_temp_file,"dump.tmp");
  grab_mem();

#ifdef LINUX
  if(ioperm(0x61,1,1))
  {
	printf("Can't open PC internal speaker i/o port.\n");
  }
  if(ioperm(0x388,2,1))
  {
	printf("Can't open adlib i/o ports.\n");
	UseAdlib=adlib_present=0;
  }
  else
  {
	UseAdlib=adlib_present=InitAdlib();
  }
  setgid(getgid());
  setuid(getuid());
#else
  UseAdlib=adlib_present=0;
#endif

  quit=QUIT_RESET;
  skip_reset=0; /* 128K mode by default */
  clean_regs();printf("Still going\n");
  open_rom48();
  open_rom128_B();
  r2_1386=ROM[2][1386];
  r2_1387=ROM[2][1387];
  r1_1386=ROM[1][1386];
  r1_1387=ROM[1][1387];
  open_rom128_E();
  open_romMF128();
  make_table();
  open_z80table();
  
  parse_params(argc,argv);
  if(ULA_delay) ULA_param=ULA_delay; /* Got to copy ULA_delay */

  InitMachine();
  make_table();
  printf("Press F1 for help, F10 to quit.\n");

  do
  {
	Sound_off();
	switch(quit)
	{
		case QUIT_HELP:
			help_scr();
			/* quit=DONT_QUIT; is now in help_scr() */
			break;
		case QUIT_NMI:
			iff2=iff1;
			iff1=halt=0;
			INC_R_REG;
			push(pc.W);
			pc.W=102;
			discard=in(0,191);
			t_state+=11;
			quit=DONT_QUIT;
			break;
		case QUIT_RESET:
			switch(reset_spectrum(skip_reset))
			{
				case 0:
					clean_regs();
					mode(128);
					joy_type=0;
					curs_caps=1;
					SLT=0;
					break;
				case 1:
					clean_regs();
					mode(48);
					joy_type=0;
					curs_caps=1;
					SLT=0;
					break;
				case 2:
					mode(128);
					break;
				case 3:
					mode(48);
					break;
				default:
					break;
			}
			quit=DONT_QUIT;
			skip_reset=255;
			if(ret_to_help)
			{
				ret_to_help=0;
				quit=QUIT_HELP;
			}
			break;
		case QUIT_JOY:
			joystick();
			quit=DONT_QUIT;
			if(ret_to_help)
			{
				ret_to_help=0;
				quit=QUIT_HELP;
			}
			break;
		case QUIT_LOAD:
			global_index=sna_index;
			strcpy(the_path,the_sna_path);
			if(!(test_fsl(".z80",".sna",".slt")))
			{
				if(different_extension(".sna",the_path))
				{
					if(!(open_z80()))
					{
						halt=0; /* Re-enable processor for new snap */
					}
				}
				else
				{
					if(!(open_sna()))
					{
						halt=0;
						ED[77].opcode(); /* reti */
					}
				}
			}
			strcpy(the_sna_path,the_path);
			sna_index=global_index;
			quit=DONT_QUIT;
			if(ret_to_help)
			{
				ret_to_help=0;
				quit=QUIT_HELP;
			}
			break;
		case QUIT_SAVE:
			strcpy(the_path,the_sna_path);
			if(!(input_string(4,"Save Z80 File")))
			{
				save_z80();
			}
			quit=DONT_QUIT;
			if(ret_to_help)
			{
				ret_to_help=0;
				quit=QUIT_HELP;
			}
			break;
		case QUIT_TAP:
			global_index=tap_index;
			strcpy(the_path,the_tap_path);
			if(!(test_fsl(".tap",".voc","")))
			{
				if(different_extension(".voc",the_path))
				{
					TAP_select();
					TAP_rompokes();
					if(VOC_file_in_open)
					{
						VOC_file_in_open=0;
						fclose(v_in_handle);
					}
				}
				else
				{
					open_VOC_input_file();
					undo_TAP_rompokes();
				}
			}
			strcpy(the_tap_path,the_path);
			tap_index=global_index;
			quit=DONT_QUIT;
			break;
		case QUIT_VOC:
			VOC_options();
			quit=DONT_QUIT;
			break;
		case QUIT_SOUND:
			strcpy(the_path,the_psg_path);
			sound_options();
			if(!all_sound_on)
			{
				Sound_off();
				UseAdlib=0;
			}
			else
			{
				UseAdlib=adlib_present;
				Sound_on();
			}
			quit=DONT_QUIT;
			if(ret_to_help)
			{
				ret_to_help=0;
				quit=DONT_QUIT;
			}
			break;
		default:
			break;
	}
	if(quit==QUIT_HELP) continue;
	if(ULA_param)
	{
		ULA_delay=ULA_param;
		ULA_param=0;
	}
	Sound_on();
#ifndef OVERSCAN
	/*init_timer();*/
#endif
	if(slow_down)
	{
		if(bitpix==1)
		{
			main_mono_loop_delay();
		}
		else
		{
			main_8bit_loop_delay();
		}
	}
	else
	{
		if(bitpix==1)
		{
			main_mono_loop();
		}
		else
		{
			main_8bit_loop();
		}
	}
#ifndef OVERSCAN
	/*release_timer();*/
#endif
  } while (quit!=QUIT_COMPLETELY);
  if(TAP_file_open) fclose(t_handle);
  if(VOC_file_in_open) fclose(v_in_handle);
  if(PSG_open) fclose(p_handle);
  release_mem();

  TrashAdlib();
  TrashMachine();
  printf("%s\n",error_str);
  printf("ULA Delay %d\n",ULA_delay);
}
